{   Program WAV_HAL [options]
*
*   Read a WAV file and create the PIC assembler source file for the sound
*   in the format required by the HAL firmware.
*
*   Command line options are:
*
*     <input file name>
*
*     <output file name>
*
*     -FROM s
*
*       Start the output sound at S seconds from the start of the input sound.
*
*     -TO s
*
*       End the output sound S seconds from the start of the input sound.
*
*     -RRATE m
*
*       Set the relative source stream playback rate.  Values above 1 cause the
*       source stream to be played back more quickly (more samples/second) and
*       values below 1 more slowly.  The default is 1, which causes the source
*       stream to be played at its intended speed.
}
program wav_hal;
%include '(cog)lib/sys.ins.pas';
%include '(cog)lib/util.ins.pas';
%include '(cog)lib/string.ins.pas';
%include '(cog)lib/file.ins.pas';
%include '(cog)lib/math.ins.pas';

const
  filter_rad = 1.50;                   {max non-zero filter kernel radius}
  filt_nh = 50;                        {number of points in half filter function}
  rate_out = 1953;                     {output stream sample rate}
  pi = 3.14159265359;                  {what it sounds like, don't touch}
  fmt_size = 24;                       {bytes in whole "fmt " chunk for uncompressed}

  filt_n = filt_nh * 2 + 1;            {total number of points in filter function}
  filt_nlast = filt_n - 1;             {index of last filter func entry, first = 0}

type
  wav_fmt_t = record                   {data from WAV "fmt" chunk}
    size: int32u_t;                    {size of remainder of the chunk}
    data_p: univ_ptr;                  {pointer to start of raw data in memory}
    dsize: sys_int_adr_t;              {size of data in bytes}
    nsamp: sys_int_machine_t;          {number of samples in the raw data}
    data_end_p: univ_ptr;              {points to first adr past end of raw data}
    dtype: integer16;                  {data type, 1 = uncompressed}
    n_chan: int16u_t;                  {number of audio channels}
    samp_sec: int32u_t;                {number of samples frames per second}
    bytes_sec: int32u_t;               {number of bytes per second}
    bytes_samp: int16u_t;              {bytes per sample frame}
    bytes_sampp: sys_int_adr_t;        {bytes per individual sample point}
    bits_samp: int16u_t;               {bits per individual sample point}
    more_p: univ_ptr;                  {pointer to additional WAV FMT data, if any}
    end;

  filt_func_t =                        {filter kernel function table}
    array[0 .. filt_nlast] of math_xy_t; {X is output samples from center}

var
  conn_in: file_conn_t;                {connection to WAV input file}
  wav_map: file_map_handle_t;          {handle to entire WAV file mapped to memory}
  wav_fmt: wav_fmt_t;                  {info about format of WAV data}
  wav_p: univ_ptr;                     {points to start of WAV file in memory}
  wav_end_p: univ_ptr;                 {points to first address after WAV file}
  conn_out: file_conn_t;               {connection to ASPIC output file}
  conn_wout: file_conn_t;              {connection to WAV output file}
  sec_st, sec_en: real;                {input WAV seconds range to write to output}
  rrate: real;                         {relative source playback rate}
  iosm, iosb: real;                    {input to output sample number transform}
  oism, oisb: real;                    {output to input sample number transform}
  n_osamp: sys_int_machine_t;          {number of output sound samples}
  osam: sys_int_machine_t;             {output sample number}
  v: real;                             {sample value}
  maxv: real;                          {largest amplitude found, 0 to +1 scale}
  time: real;                          {output time in seconds}
  filt_func: filt_func_t;              {filter kernel convolution function table}
  obuf: string_var80_t;                {one line ASPIC file output buffer}
  modname: string_var80_t;             {ASPIC file module name}
  tk: string_var80_t;                  {scratch token}
  wordval: sys_int_machine_t;          {value of word containing two samples}
  firstime: real;                      {seconds time of first sample when HAVELOW TRUE}
  havelow: boolean;                    {TRUE if first sample in low bits of WORDVAL}
{
**********************************************************************
*
*   Subroutine OWAV (D, LEN)
*
*   Write the first LEN bytes starting at D to the WAV output file.
}
procedure owav (                       {write bytes to the WAV output file}
  in      d: univ int8u_t;             {the data to write}
  in      len: sys_int_adr_t);         {the number of bytes to write}
  val_param; internal;

var
  stat: sys_err_t;

begin
  file_write_bin (d, conn_wout, len, stat);
  sys_error_abort (stat, 'wav_hal_prog', 'err_write_wav', nil, 0);
  end;
{
**********************************************************************
*
*   Subroutine OWAV_STR (S, LEN)
*
*   Write the string S of length LEN to the WAV output file.
}
procedure owav_str (                   {write string to WAV output file}
  in      s: string;                   {the input string}
  in      len: sys_int_machine_t);     {the number of characters in the string}
  val_param; internal;

begin
  owav (s, len);                       {write string to WAV file as raw data}
  end;
{
**********************************************************************
*
*   Subroutine OWAV_8U (V)
*
*   Write the unsigned byte value V to the WAV output file.
}
procedure owav_8u (                    {write unsigned byte to WAV output file}
  in      v: int8u_t);                 {0-255 byte value to write}

begin
  owav (v, 1);
  end;
{
**********************************************************************
*
*   Subroutine OWAV_16U (V)
*
*   Write the unsigned 16 bit integer value V to the WAV output file.
}
procedure owav_16u (                   {write unsigned 16 bit int to WAV output file}
  in      v: int16u_t);                {value to write}

begin
  owav_8u (v & 255);                   {write bytes in low to high order}
  owav_8u (rshft(v, 8) & 255);
  end;
{
**********************************************************************
*
*   Subroutine OWAV_16S (V)
*
*   Write the signed 16 bit integer value V to the WAV output file.
}
procedure owav_16s (                   {write signed 16 bit int to WAV output file}
  in      v: integer16);               {value to write}

begin
  owav_8u (v & 255);                   {write bytes in low to high order}
  owav_8u (rshft(v, 8) & 255);
  end;
{
**********************************************************************
*
*   Subroutine OWAV_32U (V)
*
*   Write the unsigned 32 bit integer value V to the WAV output file.
}
procedure owav_32u (                   {write unsigned 32 bit int to WAV output file}
  in      v: int32u_t);                {value to write}

begin
  owav_8u (v & 255);                   {write bytes in low to high order}
  owav_8u (rshft(v, 8) & 255);
  owav_8u (rshft(v, 16) & 255);
  owav_8u (rshft(v, 24) & 255);
  end;
{
**********************************************************************
*
*   Subroutine WRITES
*
*   Write the string in OBUF as the next line to the output file.  OBUF is
*   reset to empty.
}
procedure writes;
  internal;

var
  stat: sys_err_t;

begin
  file_write_text (obuf, conn_out, stat); {write the string to the file}
  sys_error_abort (stat, '', '', nil, 0);
  obuf.len := 0;
  end;
{
**********************************************************************
*
*   Subroutine WRITE_SAMP (V, TIME)
*
*   Write the sample value V to the file.  This routine actually writes
*   one data word to the file every second sample, because samples are
*   packed two per data word.  The sample value V is in the -1.0 to 1.0
*   range.  It will be converted to the appropriate representation needed
*   by the embedded system.  TIME is the output time of this sample
*   in seconds.
}
procedure write_samp (                 {write one sample}
  in      v: real;                     {-1.0 to 1.0 sample value}
  in      time: real);                 {output time of sample in seconds}
  val_param; internal;

var
  ival: sys_int_machine_t;             {0 to 127 embedded system sample value}
  samps: string_var16_t;               {samples string for output file}
  tk: string_var80_t;                  {scratch string}

begin
  tk.max := size_char(tk.str);         {init local var string}

  samps.max := size_char(samps.str);   {init local var string}

  ival := trunc((v / 2.0 + 0.5) * 128.0); {make raw integer sample value}
  ival := max(0, min(127, ival));      {clip to valid range}

  if not havelow then begin            {starting a new output word ?}
    wordval := ival;                   {init the output word with this sample}
    firstime := time;                  {save time of fist sample this word}
    havelow := true;                   {indicate a sample is in low bits of WORDVAL}
    return;                            {done with this sample for now}
    end;
{
*   The previous sample is in the low 7 bits of WORDVAL.  Pack the new
*   sample into the next 7 bits and write one DATA directive to the output
*   file.
}
  wordval := lshft(ival, 7) ! wordval; {pack second sample into WORDVAL}
  string_f_int16h (samps, wordval);    {make hex string of data value to write}

  string_vstring (obuf, ' data h'''(0), -1); {build the output line in OBUF}
  string_append (obuf, samps);
  string_append1 (obuf, '''');
  string_appends (obuf, '  ;'(0));
  string_f_fp_ftn (tk, firstime, 8, 3);
  string_append (obuf, tk);
  writes;                              {write the output line to the output file}

  havelow := false;                    {indicate WORDVAL is empty}
  end;
{
******************************************************************************
*
*   Subroutine OSAMP (V, TIME)
*
*   Write an output sample to all the audio output files.  V is the sample
*   value in the -1.0 to +1.0 range.  TIME is the output time in seconds of
*   this sample.
}
procedure osamp (                      {write output sample}
  in      v: real;                     {-1 to +1 sample value}
  in      time: real);                 {output time in seconds of this sample}
  val_param; internal;

var
  ival: sys_int_machine_t;             {0 - 255 WAV file unsigned byte sample value}

begin
  write_samp (v, time);                {write the sample to the ASPIC output file}

  ival := trunc(max(0.0, min(255.9, (v + 1.0) * 128.0))); {make 0-255 byte value}
  owav_8u (ival);                      {write the byte to the WAV output file}
  end;
{
******************************************************************************
*
*   Function WAV_BYTE (P, END_P)
*
*   Returns the WAV file 0-255 byte memory mapped at address P.  P will be
*   incrememted to point to the next byte.  END_P is the first address after
*   the WAV file in memory.  This routine will bomb with an error if the
*   requested byte is not within the WAV file.
}
function wav_byte (                    {get WAV file 0-255 byte value}
  in out  p: univ_ptr;                 {pointer to byte, will be incremented}
  in      end_p: univ_ptr):            {points to first address past WAV file end}
  int8u_t;
  val_param; internal;

var
  byte_p: ^int8u_t;                    {pointer to the byte}

begin
  if sys_int_adr_t(p) >= sys_int_adr_t(end_p) then begin {past end of file ?}
    sys_message_bomb ('wav_hal_prog', 'wav_eof', nil, 0);
    end;

  byte_p := p;                         {copy address to pointer of the right type}
  wav_byte := byte_p^;                 {fetch and pass back the byte value}

  p := univ_ptr(succ(byte_p));         {update P to point to the next byte}
  end;
{
******************************************************************************
*
*   Function WAV_INT16 (P, END_P)
*
*   Returns the 16 bit unsigned integer value of the next two WAV file bytes
*   at P.  END_P must point to the first address after where the end of the
*   WAV file is mapped into memory.  This routine takes care of byte ordering
*   issues.
}
function wav_int16 (                   {get WAV file 16 bit unsigned integer value}
  in out  p: univ_ptr;                 {pointer to value, will be incremented}
  in      end_p: univ_ptr):            {points to first address past WAV file end}
  int16u_t;
  val_param; internal;

var
  v: int16u_t;

begin
  v := wav_byte (p, end_p);            {get low byte}
  v := v ! lshft(wav_byte(p, end_p), 8); {get high byte}
  wav_int16 := v;                      {pass back the result}
  end;
{
******************************************************************************
*
*   Function WAV_INT16S (P, END_P)
*
*   Returns the 16 bit signed integer value of the next two WAV file bytes
*   at P.  END_P must point to the first address after where the end of the
*   WAV file is mapped into memory.  This routine takes care of byte ordering
*   issues.
}
function wav_int16s (                  {get WAV file 16 bit signed integer value}
  in out  p: univ_ptr;                 {pointer to value, will be incremented}
  in      end_p: univ_ptr):            {points to first address past WAV file end}
  integer16;
  val_param; internal;

var
  v: integer16;

begin
  v := wav_byte (p, end_p);            {get low byte}
  v := v ! lshft(wav_byte(p, end_p), 8); {get high byte}
  wav_int16s := v;                     {pass back the result}
  end;
{
******************************************************************************
*
*   Function WAV_INT32 (P, END_P)
*
*   Returns the 32 bit unsigned integer value of the next four WAV file bytes
*   at P.  END_P must point to the first address after where the end of the
*   WAV file is mapped into memory.  This routine takes care of byte ordering
*   issues.
}
function wav_int32 (                   {get WAV file 32 bit unsigned integer value}
  in out  p: univ_ptr;                 {pointer to value, will be incremented}
  in      end_p: univ_ptr):            {points to first address past WAV file end}
  int32u_t;
  val_param; internal;

var
  v: int32u_t;

begin
  v := wav_byte (p, end_p);            {get low byte}
  v := v ! lshft(wav_byte(p, end_p), 8);
  v := v ! lshft(wav_byte(p, end_p), 16);
  v := v ! lshft(wav_byte(p, end_p), 24); {get high byte}
  wav_int32 := v;                      {pass back the result}
  end;
{
******************************************************************************
*
*   Subroutine WAV_STR (P, END_P, N, VSTR)
*
*   Return the N characters starting at P into the varstring VSTR.
}
procedure wav_str (                    {get fixed length string from WAV file}
  in out  p: univ_ptr;                 {pointer to value, will be incremented}
  in      end_p: univ_ptr;             {points to first address past WAV file end}
  in      n: sys_int_machine_t;        {number of characters in the string}
  in out  vstr: univ string_var_arg_t); {the returned string}
  val_param; internal;

var
  i: sys_int_machine_t;                {loop counter}

begin
  vstr.len := 0;

  for i := 1 to n do begin             {once for each character in source string}
    string_append1 (vstr, chr(wav_byte(p, end_p)));
    end;
  end;
{
******************************************************************************
*
*   Function WAV_SAMP_MONO (N)
*
*   Returns the -1 to +1 monophonic value of the WAVE level of sample N.
*   Sample numbering starts at 0.  Zero is returned for non-existant samples,
*   meaning N is less than 0 or greater than the number of the last sample.
}
function wav_samp_mono (               {return -1 to +1 monophonic value of sample N}
  in      n: sys_int_machine_t):       {0-N number of sample to return value of}
  real;                                {-1.0 to +1.0 value of sample}
  val_param; internal;

var
  p: univ_ptr;                         {pointer to start of selected sample data}
  i, j: sys_int_machine_t;             {loop counters and scratch integers}
  acc: sys_int_conv32_t;               {sum of all channels, each chan INT16U}
  chan: sys_int_conv32_t;              {channel value normalized to 16 bits}

begin
  wav_samp_mono := 0.0;                {init sample value for non-existant case}
  if n < 0 then return;                {before start of data ?}
  if n >= wav_fmt.nsamp then return;   {after end of data ?}

  p := univ_ptr(                       {set pointer to start of selected sample}
    sys_int_adr_t(wav_fmt.data_p) + n * wav_fmt.bytes_samp);
  acc := 0;                            {init sum of all channels}

  for i := 1 to wav_fmt.n_chan do begin {once for each audio channel}
    if wav_fmt.bytes_sampp = 1
      then begin                       {sample point is single unsigned byte}
        chan := lshft(wav_byte(p, wav_fmt.data_end_p), 8);
        end
      else begin                       {multiple bytes forming signed integer}
        for j := 1 to wav_fmt.bytes_sampp do begin {once for each byte in sample}
          chan := rshft(chan, 8);      {make room for the new byte}
          chan := chan ! lshft(wav_byte(p, wav_fmt.data_end_p), 8); {merge in byte}
          end;                         {back for next byte in this sample point}
        chan := chan + 32768;          {convert to unsigned integer sample value}
        chan := chan & 16#FFFF;
        end
      ;
{
*   The unsigned integer value of this sample point is in the high bits of CHAN.
*
*   Now replicate the high sample bits into the low bits until all 16 bits
*   are filled with valid data.  This results in a sample value with the range
*   of 0 to 65535 regardless of the original format.
}
    j := wav_fmt.bits_samp;            {init number of valid bits currently in CHAN}
    while j < 16 do begin              {loop until high bits replicated into all low}
      chan := chan ! rshft(chan, j);   {replicate high J bits into next J bits}
      j := j + j;                      {update number of valid bits now in CHAN}
      end;                             {back until all 16 CHAN bits filled}

    chan := chan - 32768;              {convert to signed 16 bit integer value}
    acc := acc + chan;                 {add this channel contribution into ACC}
    end;                               {back to get value from next channel}

  wav_samp_mono :=                     {pass back final -1.0 to +1.0 frame value}
    acc / (wav_fmt.n_chan * 32768.0);
  end;
{
******************************************************************************
*
*   Function OUT_SAMP (N)
*
*   Returns the -1 to +1 value of the output sample N.  The output sample
*   numbers start at 0.  The samples written to the output file are therefore
*   numbered 0 thru N_OSAMP-1.
}
function out_samp (                    {make output sample}
  in      n: sys_int_machine_t)        {0 thru N_OSAMP-1 output sample number}
  :real;                               {-1 to +1 output sample value}
  val_param; internal;

var
  isam: sys_int_machine_t;             {input sample number}
  iinc: sys_int_machine_t;             {increment for next input sample number}
  nisam: sys_int_machine_t;            {number of input samples to filter}
  i: sys_int_machine_t;                {scratch integer and loop counter}
  ist, ien: real;                      {starting and ending input sample numbers}
  acc: real;                           {filtered value accumulator}
  wacc: real;                          {summed weight of accumulatd filtered values}
  v: real;                             {scratch data value}
  w: real;                             {scratch filter weight}
  x: real;                             {filter function X coordinate}

begin
  ist := (n + filt_func[0].x) * oism + oisb; {starting input sample number}
  ien := (n + filt_func[filt_nlast].x) * oism + oisb; {ending input sample number}

  if oism >= 0.0
    then begin                         {input/output ranges in same direction}
      iinc := 1;                       {input sample increment}
      isam := trunc(ist) + 1;          {starting input sample number}
      i := trunc(ien);                 {ending input sample number}
      nisam := max(1, i - isam + 1);   {number of input samples to filter}
      end
    else begin                         {input/output range is flipped}
      iinc := -1;                      {input sample increment}
      isam := trunc(ist);              {starting input sample number}
      i := trunc(ien) + 1;             {ending input sample number}
      nisam := max(1, isam - i + 1);   {number of input samples to filter}
      end
    ;
{
*   The following state has been set:
*
*   ISAM  -  starting input sample number
*
*   NISAM  -  number of input samples to filter, guaranteed to be at least 1
*
*   IINC  -  increment to make next input sample number
}
  acc := 0.0;                          {init filtered values accumulator}
  wacc := 0.0;                         {init weight of accumulated filtered values}

  for i := 1 to nisam do begin         {once for each input sample to filter}
    v := wav_samp_mono (isam);         {get the value of this input sample}
    x := isam * iosm + iosb;           {make output sample number at this in sample}
    x := x - n;                        {make offset from filter center}
    w := math_ipolate (filt_func, filt_n, x); {get filter weight here}
    acc := acc + v * w;                {add weighted value to accumulator}
    wacc := wacc + abs(w);             {update total filter weight}
    isam := isam + iinc;               {advance to next input sample}
    end;                               {back to process next input sample to filter}

  out_samp := acc / wacc;              {pass back filtered value}
  end;
{
******************************************************************************
*
*   Subroutine SET_FILT
*
*   Set up the filter kernel function.  This function is defined by the
*   table FILT_FUNC.  This function table contains FILT_N X,Y value pairs.
*   For each pair, X is the distance from the output sample in units of
*   output samples.  Y is the relative weighting factor of the filter
*   convolution kernel at X.  The scale of the Y values does not matter
*   as long as the kernel has a non-zero area.
}
procedure set_filt;                    {set up filter function kernel in FILT_FUNC}
  internal;

const
  maxd = filter_rad;                   {max distance at which function is non-zero}

var
  i: sys_int_machine_t;                {FILT_FUNC index loop counter}
  x: real;                             {distance from output sample at this index}

begin
  for i := 0 to filt_nlast do begin    {once for each table entry to fill in}
    x := (i - filt_nh) * maxd / filt_nh; {displacement from output sample point}
    filt_func[i].x := x;               {set function input value here}
    filt_func[i].y :=                  {filter kernel weight here}
      sqr(cos(x * 0.5 * pi / maxd));
    end;
  end;
{
******************************************************************************
*
*   Subroutine OPEN_WAV (FNAM, CONN, MAP_H, START_P, END_P, FMT)
*
*   Open a WAV file and map its contents into memory.  FNAM is the WAV file
*   name.  CONN will be returned as the connection to the open WAV file.
*   MAP_H will be the handle to the entire WAV file mapped to memory.
*   START_P will be set to the memory address of the start of the WAV file.
*   DATA_P will be set to the starting memory address of the raw data
*   in the WAV file.  FMT will be set indicating the format of the
*   raw data in the WAV file.
*
*   This routine is specific to this program.  It bombs on error instead
*   of returning a status code.
}
procedure open_wav (                   {open WAV file for memory mapped read access}
  in      fnam: string_treename_t;     {WAV file name, ".wav" suffix will be assumed}
  out     conn: file_conn_t;           {returned connection to WAV file}
  out     map_h: file_map_handle_t;    {handle to entire WAV file mapped to memory}
  out     start_p: univ_ptr;           {pointer to start of WAV file in memory}
  out     end_p: univ_ptr;             {points for first address after WAV file}
  out     fmt: wav_fmt_t);             {information about the raw data format}
  val_param; internal;

const
  max_msg_parms = 2;                   {max parameters we can pass to a message}

var
  finfo: file_info_t;                  {file system info about the WAV file}
  olen: sys_int_adr_t;                 {amount of data actually transferred}
  p: univ_ptr;                         {current WAV file reading pointer}
  chname: string_var4_t;               {name of current chunk}
  adr: sys_int_adr_t;                  {scratch address}
  pick: sys_int_machine_t;             {number of token picked from list}
  in_riff: boolean;                    {within outer RIFF chunk}
  in_wave: boolean;                    {within WAVE chunk}
  msg_parm:                            {parameter references for messages}
    array[1..max_msg_parms] of sys_parm_msg_t;
  stat: sys_err_t;                     {completion status code}

label
  loop_chunk, done_chunks;

begin
  chname.max := size_char(chname.str); {init local var string}

  file_open_map (                      {open WAV file for memory mapped access}
    fnam, '.wav',                      {input file name and mandatory suffix}
    [file_rw_read_k],                  {open of read access only}
    conn,                              {returned connection to the file}
    stat);
  sys_error_abort (stat, 'wav_hal_prog', 'wav_map_open_read', nil, 0);

  file_info (                          {get the length of this file}
    conn.tnam,                         {name of file to get info about}
    [file_iflag_len_k],                {type of info being requested}
    finfo,                             {the returned file info}
    stat);
  sys_error_abort (stat, 'wav_hal_prog', 'wav_get_length', nil, 0);

  file_map (                           {map the entire WAV file into memory}
    conn,                              {connection to the file}
    0,                                 {file offset for start of region to map}
    finfo.len,                         {length of region to map (entire file)}
    [file_rw_read_k],                  {requesting read-only access}
    start_p,                           {returned pointer to start of mapped region}
    olen,                              {length actually mapped}
    map_h,                             {handle to the mapped region}
    stat);
  sys_msg_parm_vstr (msg_parm[1], conn.tnam);
  sys_msg_parm_int (msg_parm[2], finfo.len);
  sys_error_abort (stat, 'wav_hal_prog', 'wav_map_read', msg_parm, 2);

  end_p := univ_ptr(                   {make first address past end of WAV file data}
    sys_int_adr_t(start_p) + finfo.len);
{
*   The WAV file has been opened and mapped to memory starting at START_P.
*
*   Now wade thru the headers and process the WAV subchunks until the "data"
*   subchunk is encountered.
}
  in_riff := false;                    {init to not within RIFF chunk}
  in_wave := false;                    {init to not within WAVE chunk}
  wav_fmt.size := 0;                   {init to no "fmt" chunk found yet}
  p := start_p;                        {init WAV read pointer to start of file}

loop_chunk:                            {back here to read each new chunk}
  wav_str (p, end_p, 4, chname);       {get chunk name}

  if
      (not in_riff) and                {at start of file ?}
      (not string_equal (chname, string_v('RIFF'))) {top chunk is not RIFF ?}
      then begin
    sys_message_bomb ('wav_hal_prog', 'not_riff', nil, 0);
    end;

  string_unpad (chname);               {remove any trailing spaces}
  string_tkpick80 (chname,             {pick chunk name from list}
    'RIFF LIST WAVE fmt data',
    pick);
  case pick of
{
*   RIFF chunk.  This must be the outer chunk, and is only allowed as the
*   first chunk.
}
1: begin
  if in_riff then begin
    sys_message_bomb ('wav_hal_prog', 'riff_riff', nil, 0);
    end;
  in_riff := true;

  discard( wav_int32 (p, end_p) );     {skip over length word}
  end;
{
*   LIST chunk.  The wave chunk we want could be within one of these.
}
2: begin
  discard( wav_int32 (p, end_p) );     {skip over length word}
  end;
{
*   WAVE chunk.  This has no data of its own, not even a length word.
}
3: begin
  if in_wave then begin
    sys_message_bomb ('wav_hal_prog', 'wave_wave', nil, 0);
    end;
  in_wave := true;
  end;
{
*   "fmt" chunk.  This describes the format of WAVE data.
}
4: begin
  if not in_wave then begin
    sys_msg_parm_vstr (msg_parm[1], chname);
    sys_message_bomb ('wav_hal_prog', 'not_in_wave', nil, 0);
    end;

  wav_fmt.size := wav_int32 (p, end_p); {get remaining length of this chunk}
  if odd(wav_fmt.size) then wav_fmt.size := wav_fmt.size + 1; {round up to even len}
  adr := sys_int_adr_t(p) + wav_fmt.size; {make first address after this chunk}

  wav_fmt.dtype := wav_int16s (p, end_p);
  wav_fmt.n_chan := wav_int16 (p, end_p);
  wav_fmt.samp_sec := wav_int32 (p, end_p);
  wav_fmt.bytes_sec := wav_int32 (p, end_p);
  wav_fmt.bytes_samp := wav_int16 (p, end_p);
  wav_fmt.bits_samp := wav_int16 (p, end_p);

  fmt.more_p := p;                     {save pointer to additional FMT data, if any}
  p := univ_ptr(adr);                  {read next byte immediately after chunk}
  end;
{
*   "data" chunk.  Starts with length word, then contains the raw WAV data.
}
5: begin
  if not in_wave then begin
    sys_msg_parm_vstr (msg_parm[1], chname);
    sys_message_bomb ('wav_hal_prog', 'not_in_wave', nil, 0);
    end;
  if fmt.size = 0 then begin
    sys_message_bomb ('wav_hal_prog', 'data_no_fmt', nil, 0);
    end;

  fmt.dsize := wav_int32 (p, end_p);   {get remaining length of this chunk}
  fmt.data_p := p;                     {save pointer to start of raw data}
  fmt.data_end_p := univ_ptr(          {set pointer to first adr after data end}
    sys_int_adr_t(p) + fmt.dsize);
  fmt.nsamp := fmt.dsize div fmt.bytes_samp; {number of whole samples in the data}
  fmt.bytes_sampp := (fmt.bits_samp + 7) div 8;

  goto done_chunks;                    {done looking thru chunks}
  end;
{
*   Unrecognized chunk types.  These will be assumed to have a length word
*   following the chunk name, and will be skipped over entirely.
}
otherwise                              {unrecognized chunk type}
    adr := wav_int32 (p, end_p);       {get size of this chunk}
    if odd(adr) then adr := adr + 1;   {round up to even chunk size}
    p := univ_ptr(                     {jump to first address after this chunk}
      sys_int_adr_t(p) + adr);
    end;
  goto loop_chunk;                     {back to process next chunk}

done_chunks:                           {done processing chunks, data chunk found}
  end;
{
******************************************************************************
*
*   Subroutine CMLINE
*
*   Read the command line and process the information.  The program will be
*   aborted if any errors are detected.  The following variables will be
*   set when this routine returns:
*
*     CONN_IN  -  Connection to the WAV input file mapped to memory.
*
*     MAP_WAV  -  Handle to entire WAV file mapped to memory.
*
*     WAV_FMT  -  Format information for the WAV data.
*
*     WAV_DATA_P  -  Pointer to where start of WAV data is mapped to memory.
*
*     SEC_ST, SEC_EN  -  Starting and ending seconds from the input sound
*       that is to be written to the output file.  The input sound starts
*       at 0 seconds.
*
*     CONN_OUT  -  Connection to the ASPIC output file open for text write.
}
procedure cmline;                      {read command line and process the info}
  internal;

const
  max_msg_args = 2;                    {max arguments we can pass to a message}

var
  iname_set: boolean;                  {input file name has been set}
  oname_set: boolean;                  {output file name has been set}
  fnam_in, fnam_out: string_treename_t; {input and output file names from cmd line}
  sec: real;                           {seconds play time of input sound}

  opt: string_treename_t;              {upcased command line option}
  parm: string_treename_t;             {command line option parameter}
  pick: sys_int_machine_t;             {number of token picked from list}
  msg_parm:                            {references arguments passed to a message}
    array[1..max_msg_args] of sys_parm_msg_t;
  stat: sys_err_t;                     {completion status code}

label
  next_opt, err_parm, parm_bad, done_opts;

begin
  opt.max := size_char(opt.str);       {init local var strings}
  parm.max := size_char(parm.str);
  fnam_in.max := size_char(fnam_in.str);
  fnam_out.max := size_char(fnam_out.str);
{
*   Initialize our state before reading the command line options.
}
  iname_set := false;                  {no input file name specified}
  oname_set := false;                  {no output file name specified}
  sec_st := 0.0;                       {start output sound at beginning of input}
  sec_en := 1.0e35;                    {end output sound at end of input}
  rrate := 1.0;                        {init to normal source playback rate}

  string_cmline_init;                  {init for reading the command line}
{
*   Back here each new command line option.
}
next_opt:
  string_cmline_token (opt, stat);     {get next command line option name}
  if string_eos(stat) then goto done_opts; {exhausted command line ?}
  sys_error_abort (stat, 'string', 'cmline_opt_err', nil, 0);
  if (opt.len >= 1) and (opt.str[1] <> '-') then begin {implicit pathname token ?}
    if not iname_set then begin        {input file name not set yet ?}
      string_copy (opt, fnam_in);      {set input file name}
      iname_set := true;               {input file name is now set}
      goto next_opt;
      end;
    if not oname_set then begin        {output file name not set yet ?}
      string_copy (opt, fnam_out);     {set output file name}
      oname_set := true;               {output file name is now set}
      goto next_opt;
      end;
    sys_msg_parm_vstr (msg_parm[1], opt);
    sys_message_bomb ('string', 'cmline_opt_conflict', msg_parm, 1);
    end;
  string_upcase (opt);                 {make upper case for matching list}
  string_tkpick80 (opt,                {pick command line option name from list}
    '-FROM -TO -RRATE',
    pick);                             {number of keyword picked from list}
  case pick of                         {do routine for specific option}
{
*   -FROM s
}
1: begin
  string_cmline_token_fpm (sec_st, stat);
  end;
{
*   -TO s
}
2: begin
  string_cmline_token_fpm (sec_en, stat);
  end;
{
*   -RRATE r
}
3: begin
  string_cmline_token_fpm (rrate, stat);
  end;
{
*   Unrecognized command line option.
}
otherwise
    string_cmline_opt_bad;             {unrecognized command line option}
    end;                               {end of command line option case statement}

err_parm:                              {jump here on error with parameter}
  string_cmline_parm_check (stat, opt); {check for bad command line option parameter}
  goto next_opt;                       {back for next command line option}

parm_bad:                              {jump here on got illegal parameter}
  string_cmline_reuse;                 {re-read last command line token next time}
  string_cmline_token (parm, stat);    {re-read the token for the bad parameter}
  sys_msg_parm_vstr (msg_parm[1], parm);
  sys_msg_parm_vstr (msg_parm[2], opt);
  sys_message_bomb ('string', 'cmline_parm_bad', msg_parm, 2);

done_opts:                             {done with all the command line options}
{
*   All done reading the command line.
}
  if not iname_set then begin
    sys_message_bomb ('wav_hal_prog', 'no_infile', nil, 0);
    end;
{
*   Open the WAV input file and determine its data format.
}
  open_wav (                           {open WAV file, map to memory, get format}
    fnam_in, conn_in, wav_map, wav_p, wav_end_p, wav_fmt);

  if wav_fmt.dtype <> 1 then begin
    sys_msg_parm_int (msg_parm[1], wav_fmt.dtype);
    sys_message_bomb ('wav_hal_prog', 'fmt_unsup', msg_parm, 1);
    end;

  sec :=                               {seconds play time of raw input sound}
    max(1, wav_fmt.nsamp - 1) / wav_fmt.samp_sec;
  sec_st := max(0.0, min(sec, sec_st)); {clip start/end times to input start/end}
  sec_en := max(0.0, min(sec, sec_en));
{
*   Open the ASPIC output file.
}
  if not oname_set then begin          {use default output file name ?}
    string_generic_fnam (fnam_in, '.wav', fnam_out); {use generic name of input file}
    end;

  file_open_write_text (fnam_out, '.aspic', conn_out, stat); {open ASPIC output file}
  sys_error_abort (stat, '', '', nil, 0);
  string_generic_fnam (fnam_out, '.aspic', modname); {make ASPIC file module name}
{
*   Open the WAV output file.
}
  string_generic_tnam (conn_out.tnam, '.aspic', parm); {make generic treename}
  file_open_write_bin (parm, '.wav', conn_wout, stat); {open WAV output file}
  sys_error_abort (stat, '', '', nil, 0);
  end;
{
******************************************************************************
*
*   Start of main routine.
}
begin
  obuf.max := size_char(obuf.str);     {init local var strings}
  modname.max := size_char(modname.str);
  tk.max := size_char(tk.str);

  cmline;                              {read command line and open WAV input file}
{
*   The parameters from the command line have been set, and the WAV input file
*   has been mapped into memory.  WAV_FMT indicates the format and location
*   of the raw WAV input data.
*
*   Now set up static values that will be used in creating the output sound
*   stream from the input sound stream.
}
  iosm :=                              {make input to output sample number scale}
     rate_out / (wav_fmt.samp_sec * rrate);
  if sec_st > sec_en then begin        {source range going backwards ?}
    iosm := -iosm;                     {flip scale factor for inverted range}
    end;
  oism := 1.0 / iosm;                  {make output to input sample number scale}
  oisb :=                              {input sample number at output sample 0}
    sec_st * wav_fmt.samp_sec;
  iosb := -oisb * iosm;                {output sample number at input sample 0}
  n_osamp :=                           {total number of output samples}
    trunc((sec_en - sec_st) * wav_fmt.samp_sec * iosm) + 1;

  set_filt;                            {set up filter kernel function table}
{
*   Write the preamble to the ASPIC output file.
}
  havelow := false;                    {init to no partial data word buffered}

  string_vstring (obuf, ' include "hal.inc"'(0), -1); {INCLUDE directive}
  writes;
  writes;

  string_vstring (obuf, '.', 1);       {CODE directive}
  string_append (obuf, modname);
  string_appends (obuf, ' code'(0));
  writes;
  writes;

  string_copy (modname, obuf);         {table start label}
  writes;

  string_vstring (obuf, ' global '(0), -1); {GLOBAL directive}
  string_append (obuf, modname);
  writes;
  writes;

  string_f_int (tk, n_osamp);          {DATA <n samples>}
  string_vstring (obuf, ' data '(0), -1);
  string_append (obuf, tk);            {append number of samples in the sound}
  string_appends (obuf, '  ;number of samples in the sound, '(0));
  string_f_fp_fixed (tk, (n_osamp - 1) / rate_out, 2);
  string_append (obuf, tk);
  string_appends (obuf, ' seconds'(0));
  writes;
  writes;
{
*   Set up the WAV output file so that N_OSAMP output samples can be written as
*   successive bytes.
}
  owav_str ('RIFF', 4);                {RIFF chunk name}
  owav_32u (                           {bytes in remainder of whole RIFF chunk}
    4 +                                {WAVE chunk}
    fmt_size +                         {size of whole "fmt " chunk}
    8 +                                {DATA chunk name and length}
    n_osamp);                          {one data byte for each sample}

  owav_str ('WAVE', 4);                {WAVE chunk name}

  owav_str ('fmt ', 4);                {FMT chunk name}
  owav_32u (fmt_size - 8);             {size of rest of FMT chunk}
  owav_16s (1);                        {format ID, 1 = uncompressed}
  owav_16u (1);                        {number of channels}
  owav_32u (rate_out);                 {sample frames per second}
  owav_32u (rate_out);                 {bytes per second}
  owav_16u (1);                        {bytes per sample point}
  owav_16u (8);                        {bits per sampel point}

  owav_str ('data', 4);                {DATA chunk name}
  owav_32u (n_osamp);                  {size of rest of DATA chunk}
{
*   Generate all the output samples and write them to the output files.
}
  maxv := 0.0;                         {init maximum amplitude found}
  for osam := 0 to n_osamp-1 do begin  {find the maximum amplitude}
    v := out_samp (osam);              {get the value for this output sample}
    maxv := max(maxv, abs(v));         {include this sample in max amplitude}
    end;

  for osam := 0 to n_osamp-1 do begin  {once for each output sample}
    time := osam / rate_out;           {make output time in seconds}
    v := out_samp (osam);              {get the value for this output sample}
    osamp (v / maxv, time);            {write the sample value to all output files}
    end;
{
*   Done writing all the output samples.  Clean up and leave.
}
  file_close (conn_in);                {close the input file}

  if havelow then begin                {partial ASPIC data word is buffered ?}
    write_samp (0.0, 0.0);             {cause word with single sample to be written}
    end;
  writes;
  string_vstring (obuf, ' end'(0), -1); {write END directive to ASPIC output file}
  writes;

  file_close (conn_out);               {close the ASPIC output file}
  file_close (conn_wout);              {close the WAV output file}
  end.
